/* 27 March 2021, by Ruud Harmsen */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "getfhtml.h"
#include "gumbo.h"

static char *walk (GumboNode *node_p, GumboTag tag2find, char *attrname);

/* gfh_GetFromBuff expects a null-terminated buffer `buff'
   containing HTML-code. It parses the code using Google's
   gumbo-parser (see https://github.com/google/gumbo-parser).
   It then find the first occurrence of tag `tagname'.

   If `attrname' is NULL, or a pointer to an empty string,
   gfh_GetFromBuff returns the text child belonging to `tagname'.
   This is useful for example to extract the title from an
   HTML buffer.

   If `attrname' points to a non-empty string, gfh_GetFromBuff
   finds the first attribute of that name under `tagname', and
   returns the corresponding value. This is useful for example
   to find the language code specified as "<html lang=en>" or
   similar.

   Both `tagname' and `attrname' are case-insensitive.

   The return value is a pointer to a malloc'ed area, and it
   is up to the caller to free that.
   The return value is NULL if no result could be found.
 */
char *gfh_GetFromBuff (char *buff, char *tagname, char *attrname)
{
   char *retbuf = NULL;

   GumboOutput* tree = gumbo_parse(buff);

   if (tree)
   {
      GumboTag gtag = gumbo_tag_enum(tagname);
      retbuf = walk(tree->root, gumbo_tag_enum(tagname), attrname);
      gumbo_destroy_output(&kGumboDefaultOptions, tree);

      return retbuf;
   }

   return NULL;
}

static char *walk (GumboNode *node_p, GumboTag tag2find, char *attrname)
{
   char *retbuf = NULL;

   if (node_p->type == GUMBO_NODE_ELEMENT)
   {
      if (node_p->v.element.tag != tag2find)
      {
         /* Continue traversing the element tree */
         for (int i = 0; i < node_p->v.element.children.length; i++)
         {
            retbuf = walk(node_p->v.element.children.data[i], tag2find, attrname);
            if (retbuf)
               break;
         }
      }
      else
      {
         if (!attrname || !attrname[0])
         {
            /* The result to be returned is not an attribute value,
               but the text child of the tag just found. */
            if (node_p->v.element.children.length > 0)
            {
               GumboNode *textnode = node_p->v.element.children.data[0];
               if (textnode && textnode->type == GUMBO_NODE_TEXT)
               {
                  retbuf = malloc(strlen(textnode->v.text.text) + 1);
                  if (retbuf)
                  {
                     strcpy(retbuf, textnode->v.text.text);
                  }
               }
            }
            return retbuf;
         }
         else
         {
            GumboAttribute* gumatt;

            if ((gumatt = gumbo_get_attribute(&(node_p->v.element.attributes), attrname)))
            {
               retbuf = malloc(strlen(gumatt->value) + 1);
               if (retbuf)
               {
                  strcpy(retbuf, gumatt->value);
               }
            }
            return retbuf;
         }
      }
   }

   return retbuf;
}

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* gfh_GetFromHtml is similar to gfh_GetFromBuff except that
   it operates on a file instead of a buffer. It will read
   at most nbytes (not including the closing null-byte) from
   the file into an internal buffer, and call gfh_GetFromBuff
   on that buffer.

   If nbytes is zero, it means the whole file will read into
   memory and parsed.

   gfh_GetFromHtml resets the file pointer to the position it
   had before the call.
 */
char *gfh_GetFromHtml (FILE *htmlfile, size_t nbytes, char *tagname, char *attrname)
{
   char *filebuf;
   char *retbuf;
   long filepos;

   if (nbytes == 0)
   {
      struct stat filestats;

      if (fstat(fileno(htmlfile), &filestats) != 0)
      {
        fprintf(stdout, "Line=%3d in %s, returned NULL\n", __LINE__, __FILE__);
        return NULL;
      }
      else
         nbytes = filestats.st_size;
   }

   if (nbytes == 0)
     return NULL;

   if (!(filebuf = calloc(nbytes + 1, 1)))
   {
      fprintf(stdout, "Line=%3d in %s, returned NULL\n", __LINE__, __FILE__);
      return NULL;
   }
   filepos= ftell(htmlfile);
   fread(filebuf, 1, nbytes, htmlfile);
   if (ferror(htmlfile))
   {
      fprintf(stdout, "Line=%3d in %s, fread failed\n", __LINE__, __FILE__); 
      return NULL;
   }
   retbuf = gfh_GetFromBuff(filebuf, tagname, attrname);
   free(filebuf);
   fseek(htmlfile, filepos, SEEK_SET);

   return retbuf;
}


/* And the same again, now from a file pathname, null-terminated */ 
char *gfh_GetFromPath (char *path, size_t nbytes, char *tagname, char *attrname)
{
   FILE *htmlfile = fopen(path, "r");
   char *retbuf;

   if (!htmlfile)
      return NULL;
   retbuf = gfh_GetFromHtml(htmlfile, nbytes, tagname, attrname);
   fclose(htmlfile);

   return retbuf;
}

#ifdef MAIN
#include <stdio.h>
#include <string.h>

char *testtext =
"<html lang=ina-Grek\n"
"<HEAD>"
"<titLE>This is the test buffer title</title>\n"
"<h1>Header 1</H1>\n"
"</head><body>\n"
"<p><tiME dateTIME=\"2021-03-27\">27 maart 2021</time></p>\n"
"<h2>Subkopje</H2>\n"
"<p>De tekst onder het subkopje</p>\n"
"<h2>Nog een subkop</h2>\n"
"<p>Tekst behorende bij nog een subkop</p>\n"
"</body></html>\n";

int main (int argc, char **argv)
{
   char *res;
   FILE *htmlfile;

   printf("\nFinding the title\n");
   res = gfh_GetFromBuff(testtext, "TItle", NULL);
   if (res)
      printf("Title\t= \"%s\"\n", res);
   free(res);

   printf("\nFinding the language\n");
   res = gfh_GetFromBuff(testtext, "htmL", "laNg");
   if (res)
      printf("Lang\t= \"%s\"\n", res);
   free(res);

   printf("\nFinding the date\n");
   res = gfh_GetFromBuff(testtext, "Time", "datetime");
   if (res)
      printf("Pubdate\t= \"%s\"\n", res);
   free(res);

   for ( ; argc > 1; argc--, argv++)
   {
      if ((htmlfile = fopen(argv[1], "r")) != NULL)
      {
         printf("\nFinding the title in file %s\n", argv[1]);
         res = gfh_GetFromHtml(htmlfile, 1 << 10, "title", NULL);
         if (res)
            printf("Title\t= \"%s\"\n", res);
         free(res);

         printf("\nFinding the language in file %s\n", argv[1]);
         res = gfh_GetFromHtml(htmlfile, 1 <<  8, "html", "lang");
         if (res)
            printf("Lang\t= \"%s\"\n", res);
         free(res);

         printf("\nFinding the date in file %s\n", argv[1]);
         res = gfh_GetFromHtml(htmlfile, 1 << 12, "time", "datetime");
         if (res)
            printf("Pubdate\t= \"%s\"\n", res);
         free(res);

         fclose(htmlfile);
      }
   }

   return 0;
}
#endif
